// Copyright 2016, University of Colorado Boulder

/**
 *
 * @author Michael Kauzmann
 * @author Matt Pennington
 */

(function() {
  'use strict';


  // Get the initialization parameters from the json file.
  var simName = QueryStringMachine.get( 'sim', { type: 'string' } );
  window.tabs = {};

  // Load the phet-io wrapper script, see http://stackoverflow.com/questions/14521108/dynamically-load-js-inside-js
  var script = document.createElement( 'script' );
  script.src = './' + simName + '/' + simName + '.js';
  script.onload = function() {

    // Used as a constant for the 'phetID' in simIFrameClient
    var WRAPPER_NAME = 'classroomActivity';

    // W prefix is for wrapper
    var WButton = window.WrapperTypes.WButton;
    var WTextArea = window.WrapperTypes.WTextArea;
    var WRadioButton = window.WrapperTypes.WRadioButton;
    var WHTMLReport = window.WrapperTypes.WHTMLReport;

    /**
     * Persistent model object.
     *
     * @type {{currentTabId: string, activeTabId: string, tabIds: string[], predict: {initialized: boolean, prediction: string, explanation: string, simState: {}}, explore: {initialized: boolean, notes: string, simState: {}}, experiment: {initialized: boolean, currentIdeaIndex: number, ideas: Array, simState: {}}, reflect: {initialized: boolean, simState: {}, dataTable: null, rules: string, active: boolean}, apply: {initialized: boolean, simState: {}, rules: string, startedToTestCases: boolean, readyToTestCases: boolean, casesComplete: Set, cases: Array}}}
     */
    window.classroom = {
      currentTabId: 'predict',
      activeTabId: '',
      tabIds: [
        'predict',
        'explore',
        'experiment',
        'reflect',
        'apply'
      ],
      predict: {
        initialized: true,
        prediction: '', // Radio button selection, {'red'|'blue'|'tie'}
        explanation: '', // textarea user input
        simState: {}
      },
      explore: {
        initialized: false,
        notes: '', // textarea user input
        simState: {}
      },
      experiment: {
        initialized: false,
        currentIdeaIndex: -1,
        isRecordDisabled: true,
        firstTestCompleted: false,
        ideas: [
          // Sample idea object,
          // {
          //   idea: '', // idea text
          //   summary: '', // what I learned text
          //   tested: boolean,
          //   whatILearnedSubmitted: boolean  // submitting 'what i learned' doesn't deactivate the sim like 'tested' does
          //   simState: {},
          //   trials: [ // may be handled by dataTable???
          //     {
          //       id: 1,
          //       simState: {},
          //       screenshot: {},
          //       values: {},
          //       sumOfForces: 0,
          //       result: {red|blue|tie},
          //       timeToWin: 0
          //     }, ...
          //   ]
          // }, ...
        ],
        simState: {}
      },
      reflect: {
        initialized: false,
        simState: {},
        dataTable: null,
        rules: '',
        active: true,
        firstTimeSaving: true
      },
      apply: {
        initialized: false,
        simState: {},
        rules: '',
        startedToTestCases: false, // There needs to be a check on if you can test cases, and if you have started.
        readyToTestCases: false,
        printingReport: false,
        casesComplete: new Set(),
        cases: [
          // Sample case object,
          // {
          //   prediction: '', // From radio button choice {'red'|'blue'|'tie'}
          //   explanation: '', // Explanation of the prediction
          //   winningFlagImgSrc: '', // the src of the winning flag
        ]
      }
    };

    document.title = window.simConfiguration.pageTitle;

    // Load the css file specific to this sim.  Will define rules for the custom cells in the data table.
    var cssFile = document.createElement( 'link' );
    cssFile.setAttribute( 'rel', 'stylesheet' );
    cssFile.setAttribute( 'type', 'text/css' );
    cssFile.setAttribute( 'href', './' + simName + '/' + simName + '.css' );
    document.getElementsByTagName( 'head' )[ 0 ].appendChild( cssFile );

    // Load the PhET-iO Simulation
    var sim = WrapperUtils.getSim( 'john-travoltage', '1.4.0-phetiodev.1' );
    window.simIFrameClient = new SimIFrameClient( document.getElementById( 'simulation-iframe' ) );

    var recorder = new window.Recorder();

    // When the recorder is ready, launch the sim.
    recorder.start( function( sessionMetadata ) {
      window.simIFrameClient.launchSim( sim.URL + '&' + window.simConfiguration.queryString, {

        emitInputEvents: true,
        expressions: window.simConfiguration.customization.predict,
        phetioEventsListener: function( message ) {

          // Send the data
          recorder.recordMessage( message );
        },
        callback: function() {
          window.simLaunched = true;
        },
        simStartedMetadata: sessionMetadata
      } );
    } );

    /**
     * Draw the 'next' button for each tab
     * This is a trivial html element but is used 5 times.
     * It is created with js in case changes are needed.
     *
     * @param id - the window.classroom.tabId for the desired button
     * @returns {Element} - Black right pointing triangle over text = "Next"
     */
    function createNextButton( id ) {
      // dom elements
      var outer = document.createElement( 'div' );
      var first = document.createElement( 'span' );
      var second = document.createElement( 'span' );
      var triangleOuter = document.createElementNS( 'http://www.w3.org/2000/svg', 'svg' );
      var triangleInner = document.createElementNS( 'http://www.w3.org/2000/svg', 'polygon' );

      // define polygon object
      triangleInner.setAttribute( 'points', '0,0 0,100 100,50' );
      triangleInner.setAttribute( 'transform', 'scale(0.18)' );
      triangleInner.setAttribute( 'fill', 'black' );

      // define svg object
      triangleOuter.setAttribute( 'class', 'triangle-left' );
      triangleOuter.setAttribute( 'width', '18' );
      triangleOuter.setAttribute( 'height', '18' );
      triangleOuter.setAttribute( 'viewport', '0 0 10 10' );

      // define text object
      second.innerHTML = 'Next';

      // define wrapper div
      outer.setAttribute( 'class', 'next' );
      outer.setAttribute( 'style', 'display: none;' );
      outer.id = id + '-next';

      // package the objects
      triangleOuter.appendChild( triangleInner );
      first.appendChild( triangleOuter );
      outer.appendChild( first );
      outer.appendChild( second );

      return outer;
    }


    /**
     * Function to change the currently shown tab.  Alter the appropriate classes and changes the currentTabId in the classroom object.
     * Tab will only be switched if the new tab contains the class 'pickable'
     * The old tab will be given the class 'pickable'
     */
    function switchTab() {
      if ( this.classList.contains( 'pickable' ) ) {

        // Change the class for the tabs
        var currentTab = document.getElementById( window.classroom.currentTabId );
        this.classList.add( 'active' );
        this.classList.remove( 'pickable' );
        currentTab.classList.add( 'pickable' );
        currentTab.classList.remove( 'active' );

        // Switch the visible workspace
        var currentWorkspace = document.getElementById( window.classroom.currentTabId + '-workspace' );
        var newWorkspace = document.getElementById( this.id + '-workspace' );
        $( currentWorkspace ).hide();
        $( newWorkspace ).show();

        // Clear the data table
        $( '#data-table' ).innerHTML = '';

        var newTabId = this.id;

        // Get the current state, save the state in the old tab
        window.simIFrameClient.invoke( 'phetio', 'getState', [], function( oldState ) {
          // save the old state
          window.classroom[ window.classroom.currentTabId ].simState = oldState;
          window.tabs[ window.classroom.currentTabId ].onLeave();
          // update the name of the current tab
          window.classroom.currentTabId = newTabId;

          window.simIFrameClient.invoke( window.simConfiguration.modelTandem, 'reset', [], function() {
            // if this is the first time a tab has loaded, initialize it
            if ( !window.classroom[ newTabId ].initialized ) {
              window.classroom[ newTabId ].initialized = true;
              window.classroom.activeTabId = newTabId;
              window.tabs[ newTabId ].initialize();
              window.simIFrameClient.invokeSequence( window.simConfiguration.customization[ newTabId ] );
            }
            // if the tab has been loaded before, reload the state of the sim and the tab
            else {
              window.tabs[ newTabId ].onLoad();

              // Reload the customization and the state
              window.simIFrameClient.invokeSequence( window.simConfiguration.customization[ newTabId ], function() {
                window.simIFrameClient.invoke( 'phetio', 'setState', [ window.classroom[ newTabId ].simState ], function() {
                  // if this is not the active tab, interaction should be disabled
                  if ( newTabId !== window.classroom.activeTabId ) {
                    window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setInactiveAndNotInteractive,
                      window.tabs[ newTabId ].afterStateLoad );
                  }
                } );
              } );
            }
          } );
        } );

        // Record button event to the events stream
        window.recordButtonPressed( this );
      }
    }

    // Initialize tab switching behavior
    window.classroom.tabIds.forEach( function( tabName, i ) {
      // Add click listeners for tabs
      $( '#' + tabName ).click( switchTab );

      // Add invisible next buttons to the first 4 tabs
      if ( i <= 3 ) {
        document.getElementById( tabName + '-workspace' ).appendChild( createNextButton( tabName ) );
        var $next = $( '#' + tabName + '-next' );
        $next.hide();
        $next.click( function() {
          $( '#' + window.classroom.tabIds[ i + 1 ] ).click();
          // Don't record button press because it is done in the tab click listener
        } );
      }
    } );

    /**
     * Create a radio button group for choosing a hypothesis, used on the predict and apply tabs.
     *
     * @param container - HTML element to hold the radio group
     * @param radioData - [ { value: String, text: String } ] - To be given to the individual radio buttons' value and text
     * @param uniqueID - string to make the radio group have unique ids. This is also the name of the radio buttons
     * @callback callback
     */
    window.createHypothesisRadioGroup = function( container, radioData, uniqueID, callback ) {

      window.assert( !$( 'input[name="' + uniqueID + '"]' )[ 0 ], 'Unique ID is already in use' );

      for ( var i in radioData ) {
        if ( radioData.hasOwnProperty( i ) ) {
          if ( radioData[ i ].hasOwnProperty( 'value' ) && radioData[ i ].hasOwnProperty( 'text' ) ) {
            var input = document.createElement( 'input' );
            input.id = uniqueID + '-' + radioData[ i ].value;
            input.type = 'radio';
            input.value = radioData[ i ].value;
            input.name = uniqueID;

            var label = document.createElement( 'label' );
            label.setAttribute( 'for', input.id );
            label.innerHTML = radioData[ i ].text;

            container.appendChild( input );
            container.appendChild( label );
            container.appendChild( document.createElement( 'br' ) );
          }
        }
      }

      callback && callback();
    };

    /**
     * Record when a text area changes to the PhETio event stream
     * @param textArea - A text area html element
     * @param previousNotesText - String
     * @returns {string} - New 'previousNotesText'
     */
    window.recordTextAreaChanged = function( textArea, previousNotesText ) {
      var currentNotesText = textArea.value;

      // Only send messages when the notes change
      if ( currentNotesText !== previousNotesText ) {

        window.simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: WRAPPER_NAME + '.' + textArea.id,
          componentType: WTextArea,
          event: 'changed',
          parameters: {
            // Don't send the whole classroom for the textareas
            text: currentNotesText
          }
        } ] );
        return currentNotesText;
      }
      else {
        return previousNotesText;
      }
    };

    /**
     * Record when a button is pressed to the PhETio event stream
     * @param button
     */
    window.recordButtonPressed = function( button ) {
      window.simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: WRAPPER_NAME + '.' + button.id,
        componentType: WButton,
        event: 'pressed',
        parameters: {
          classroom: window.classroom
        }
      } ] );
    };

    /**
     * Record when a radio button is fired to the PhETio event stream
     * @param radioButton
     */
    window.recordRadioButtonFired = function( radioButton ) {
      window.simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: WRAPPER_NAME + '.' + radioButton.id,
        componentType: WRadioButton,
        event: 'fired',
        parameters: {
          classroom: window.classroom
        }
      } ] );
    };

    /**
     * Record the generated HTML report on final submit to the PhETio event stream
     * @param htmlReport {String}
     */
    window.recordHTMLReport = function( htmlReport ) {
      window.simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: WRAPPER_NAME + '.' + 'htmlReport',
        componentType: WHTMLReport,
        event: 'generated',
        parameters: {
          classroom: window.classroom,
          htmlReport: htmlReport
        }
      } ] );
    };

    /**
     * Draw the data table given an object
     * @param dt {DataTable} - see common/dataTable.js
     * @param whichTable {'blankExperiment'|'experiment'|'reflect'} -
     */
    window.drawDataTable = function( dt, whichTable ) {
      var tableElement = document.getElementById( 'data-table' );
      tableElement.innerHTML = '';
      $( tableElement ).show();

      // Make the record button
      var recordButton = document.createElement( 'button' );
      recordButton.id = 'phet-data-table-record';
      recordButton.classList.add( 'phet-button' );
      recordButton.type = 'button';
      recordButton.innerHTML = 'Save Trial';

      // This is for the blank table drawn at the beginning of the experiment table
      if ( whichTable === 'blankExperiment' ) {
        // Don't add the record button
      }
      // Put the record button inside the idea on the left
      else if ( whichTable === 'experiment' ) {
        var recordingSpan = document.getElementById( 'recording-div-' + (window.classroom.experiment.currentIdeaIndex + 1 ) );
        recordingSpan.appendChild( recordButton );

      }
      // Handle as else but it ONLY occurs in reflect tab or things break
      else if ( whichTable === 'reflect' ) {
        // When experiment isn't the active tab
        if ( window.classroom.activeTabId !== 'reflect' ) {
          recordButton.classList.add( 'disabled' );
        }
        var reflectWorkspace = document.getElementById( 'reflect-workspace' );
        reflectWorkspace.insertBefore( recordButton, document.getElementById( 'reflect-input' ) );
      }
      else {
        throw new Error( 'Invalid table to draw' );
      }

      // Draw the table
      dt.setCellSetters( window.simConfiguration.getExperimentCellSetters( dt ) );
      dt.drawTable( tableElement );

      // Record button listener
      $( recordButton ).click( function() {
        if ( !$( this ).hasClass( 'disabled' ) ) {
          dt.record();
        }
      } );
    };


    // Initialize the first tab
    window.tabs.predict.initialize();


  };

  document.head.appendChild( script );
})();
