// Copyright 2016, University of Colorado Boulder

/**
 * Dependent on window.toast()
 *
 * To use a DataTable, the simConfiguration file must have the following objects:
 *    recordDataPoint: function( simIFrameClient, callback )
 *      + recordDataPoint will get the information from the simIFrameClient using phet-io methods
 *      + recordDataPoint will pass the data to the dataTable using callback( dataPoint )
 *    tableHeaders: [ {text:'Column Name',width:columnWidth} ],
 *      + the sum of the widths must be less than width of table
 *      + the width of the last column may work best if you set it to 0 and let the browser determine how to display it.
 *    getCellSetters: function(dataTable)
 *      + @returns [ function( trialNumber, cell, dataPoint ) ]
 *      + the returned functions should take the following arguments:
 *        - trialNumber: Number
 *        - cell: HTML element
 *        - dataPoint: data object relevant to the sim being wrapped.
 *      + the returned functions should draw an html element using the parameters provided and append it to {cell}
 *      + public functions available for settting cells provided by dataTable are:
 *        - deleteButtonCellSetter
 *        - restoreSimCellSetter
 *        - trialNumberCellSetter
 *
 *  To draw the table:
 *    + construct it using new DataTable( simConfiguration, simIFrameClient, wrapperName )
 *    + add the cellSetters using dataTable.setCellSetters( simConfiguration.getCellSetters( dataTable ) )
 *    + add it to an exiting HTML element using dataTable.drawTable( container )
 *
 *  The data table will draw its own "Record Data" button that will add rows to the table
 *
 *
 * @author Matt Pennington
 */
(function() {
  'use strict';

  /**
   * @param recordDataPoint
   * @param simIFrameClient
   * @param wrapperName - used to write messages to the phet-io data stream
   * @param tableHeaders
   * @param className - used to allow for unique css in each table
   * @constructor
   */
  window.DataTable = function( recordDataPoint, simIFrameClient, wrapperName, tableHeaders, className ) {

    var self = this;

    var WButton = window.WrapperTypes.WButton;

    self.recordedDataPoints = [];
    self.trialNumber = 1;
    // var cellSetters = []; // Commented by Andrew to pass lint

    this.setCellSetters = function( cellSetters ) {
      self.cellSetters = cellSetters;
    };

    // Load the standard css file
    var staticCssNode = document.createElement( 'link' );

    // Note: this href assumes that it is being used by a dedicated wrapper repo, see https://github.com/phetsims/phet-io-wrappers/issues/17
    staticCssNode.href = '../common/css/DataTable.css';
    staticCssNode.type = 'text/css';
    staticCssNode.rel = 'stylesheet';
    staticCssNode.media = 'screen';
    staticCssNode.title = 'dynamicLoadedSheet';
    document.getElementsByTagName( 'head' )[ 0 ].appendChild( staticCssNode );


    // Create a new css object for any rules that are dependent on the simConfiguration
    var dynamicCssNode = document.createElement( 'style' );
    dynamicCssNode.type = 'text/css';
    // dynamicCssNode.rel = 'stylesheet';
    // dynamicCssNode.media = 'screen';
    // dynamicCssNode.title = 'dynamicSheet';

    var innerHTML = '';
    for ( var i in tableHeaders ) {
      var idx = 1 + parseInt( i, 10 );
      innerHTML += '.' + className + ' td:nth-child(' + idx + '),\n' +
                   '.' + className + ' th:nth-child(' + idx + ') { \n\twidth: ' +
                   tableHeaders[ i ].width + 'px;\n}\n';
    }
    dynamicCssNode.innerHTML = innerHTML;
    document.getElementsByTagName( 'head' )[ 0 ].appendChild( dynamicCssNode );

    /**
     * Creates a font-awesome object that will be recorded by phet-io via an event listener.
     *
     * @param componentType:String - the phet-io componentType of the element
     * @param trialNumber:Number - the trial id of the component
     *
     * @returns <i> element that will need a font-awesome class added by using d.className += ' fa-subclass';
     */
    function createPhetioElement( componentType, trialNumber ) {
      var button = document.createElement( 'i' );
      button.setAttribute( 'aria-hidden', 'true' );
      button.setAttribute( 'class', 'fa' );

      button.id = componentType + trialNumber;
      button.trialNumber = trialNumber;

      return button;
    }

    /**
     * The given row will be swapped with the row in the previous position.
     *
     * @param tableRow - a row in '#trialTable'
     */
    function decrementRow( tableRow ) {
      var previous = tableRow.prev();
      tableRow.after( previous );
    }

    /**
     * Find the index of the object in userData
     * @param {number} tn
     * @returns {number}
     */
    function trialNumberToIndex( tn ) {
      assert && assert( typeof tn === 'number', 'Trial number should be a number, but it was a ' + ( typeof tn ) );
      var index = -1;
      for ( var i = 0; i < self.recordedDataPoints.length && index === -1; i++ ) {
        if ( self.recordedDataPoints[ i ].trialNumber === tn ) {
          index = i;
        }
      }
      return index;
    }

    this.trialNumberCellSetter = function( trialNumber, cell ) {
      var trialNumberLabel = document.createElement( 'p' );
      trialNumberLabel.innerHTML = ( trialNumber ).toString();
      var decrementButton = createPhetioElement( 'decrementButton', trialNumber );
      decrementButton.className += ' fa-arrow-circle-up';
      var incrementButton = createPhetioElement( 'incrementButton', trialNumber );
      incrementButton.className += ' fa-arrow-circle-down';

      cell.appendChild( decrementButton );
      cell.appendChild( trialNumberLabel );
      cell.appendChild( incrementButton );

      decrementButton.addEventListener( 'click', function() {
        //  Swap with the previous row unless we are at the top.
        var tr = $( '#trialTableRow' + this.trialNumber );
        if ( tr.prevAll().length >= 1 ) {
          decrementRow( tr );

          // Maintain order in array so that rows can be redrawn in order.
          var idx = trialNumberToIndex( this.trialNumber );
          var temp = self.recordedDataPoints[ idx ];
          self.recordedDataPoints[ idx ] = self.recordedDataPoints[ idx - 1 ];
          self.recordedDataPoints[ idx - 1 ] = temp;
        }
        else {
          //Do not record decrement press if we don't decrement
          return;
        }

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: wrapperName + '.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: self.recordedDataPoints
          }
        } ] );
      } );

      incrementButton.addEventListener( 'click', function() {
        // Swap with subsequent row unless we are at the bottom
        var tr = $( '#trialTableRow' + this.trialNumber );
        if ( tr.nextAll().length >= 1 ) {
          decrementRow( tr.next() );

          // Maintain order in array so that rows can be redrawn in order.
          var idx = trialNumberToIndex( this.trialNumber );
          var temp = self.recordedDataPoints[ idx ];
          self.recordedDataPoints[ idx ] = self.recordedDataPoints[ idx + 1 ];
          self.recordedDataPoints[ idx + 1 ] = temp;
        }
        else {
          //Do not record increment if nothing changes
          return;
        }

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: wrapperName + '.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: self.recordedDataPoints
          }
        } ] );
      } );
    };

    this.restoreSimCellSetter = function( trialNumber, cell, dataPoint, cb ) {
      var leftPictureSpan = document.createElement( 'span' );
      leftPictureSpan.appendChild( dataPoint.image );

      var restoreButton = createPhetioElement( 'restoreButton', trialNumber );
      restoreButton.className += ' fa-play-circle';

      var rightPictureSpan = document.createElement( 'span' );
      rightPictureSpan.appendChild( restoreButton );
      var pictureDiv = document.createElement( 'div' );
      pictureDiv.appendChild( leftPictureSpan );
      pictureDiv.appendChild( rightPictureSpan );
      cell.classList.add( 'screenshot' );
      cell.appendChild( pictureDiv );

      restoreButton.addEventListener( 'click', function() {
        window.toast( 'Simulation restored to Trial ' + this.trialNumber, 3000 );

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: wrapperName + '.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: self.recordedDataPoints
          }
        } ] );
        simIFrameClient.invoke( 'phetio', 'setState', [ dataPoint.state ] , cb);
      } );
    };

    this.deleteButtonCellSetter = function( trialNumber, cell ) {
      var deleteButton = createPhetioElement( 'deleteButton', trialNumber );
      deleteButton.className += ' fa-times-circle';
      cell.appendChild( deleteButton );

      deleteButton.addEventListener( 'click', function() {
        // Get trialNumber from the id string
        var index = trialNumberToIndex( this.trialNumber );

        // We expected to see an object in userData with trialNumber == trialId, but something went wrong. Bail.
        window.assert( index !== -1, 'Attempted to delete trial that does not exist' );

        // Remove the trial object from the userData array and from the HTML table
        self.recordedDataPoints.splice( index, 1 );
        $( '#trialTableRow' + this.trialNumber ).remove();


        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: wrapperName + '.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: self.recordedDataPoints
          }
        } ] );
      } );
    };

    /**
     *
     * @param container - HTML Element that becomes the direct parent of the table.
     */
    this.drawTable = function( container ) {
      // Compile table
      var table = document.createElement( 'table' );
      var thead = document.createElement( 'thead' );
      var theadTr = document.createElement( 'tr' );
      var tbody = document.createElement( 'tbody' );
      thead.appendChild( theadTr );
      table.appendChild( thead );
      table.appendChild( tbody );

      container.appendChild( table );

      // Table settings
      table.id = 'phet-data-table-' + className;
      table.classList.add( 'phet-data-table' );
      table.classList.add( className );

      // Draw headers
      for ( var i in tableHeaders ) {
        var th = document.createElement( 'th' );
        th.innerHTML = tableHeaders[ i ].text;
        theadTr.appendChild( th );
      }

      // Draw data points
      for ( var dataPointIdx in self.recordedDataPoints ) {
        var row = tbody.insertRow( -1 );
        var dp = self.recordedDataPoints[ dataPointIdx ];
        row.id = 'trialTableRow' + dp.trialNumber;
        for ( var thIdx in tableHeaders ) {
          var cell = row.insertCell( thIdx );
          self.cellSetters[ thIdx ]( dp.trialNumber, cell, dp );
        }
      }
    };

    this.record = function() {
      recordDataPoint( simIFrameClient, function( dataPoint ) {
        dataPoint.trialNumber = self.trialNumber;
        self.recordedDataPoints.push( dataPoint );
        addDataPointToTable( dataPoint );

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: wrapperName + '.recordDataButton',
          componentType: WButton,
          event: 'pressed',
          parameters: dataPoint
        } ] );
      } );
    };

    /**
     *
     * @param dataPoint
     */
    function addDataPointToTable( dataPoint ) {
      var table = $( '.' + className + ' tbody' )[ 0 ];
      var row = table.insertRow( -1 );
      row.setAttribute( 'id', 'trialTableRow' + self.trialNumber );

      for ( var i in self.cellSetters ) {
        var cell = row.insertCell( i );
        self.cellSetters[ i ]( self.trialNumber, cell, dataPoint );
      }

      // AutoScroll to bottom
      var tb = $( '.' + className + ' tbody' )[ 0 ];
      tb.scrollTop = 100 * self.recordedDataPoints.length;

      self.trialNumber++;
    }
  };
})();

